import asyncio

from thriftpy.transport import TTransportException


class TTransport:
    def __init__(self, reader):
        self._reader = reader

    @asyncio.coroutine
    def read(self, sz):
        buff = b''
        have = 0
        while have < sz:
            chunk = yield from self._reader.read(sz - have)
            if len(chunk) == 0:
                raise TTransportException(type=TTransportException.END_OF_FILE,
                                          message='TSocket read 0 bytes')
            have += len(chunk)
            buff += chunk
        return buff



import sys
import asyncio
import async_timeout

import functools

from thriftpy.thrift import TMessageType, TApplicationException
from thriftpy.transport import TTransportException

from aiothrift.transport import TTransport
from .protocol import TBinaryProtocol
from .util import args2kwargs
from .errors import ConnectionClosedError
from .log import logger


@asyncio.coroutine
def create_connection(service, address, *, protocol_cls=TBinaryProtocol,
                      timeout=None, loop=None):
    host, port = address
    reader, writer = yield from asyncio.open_connection(
        host, port, loop=loop)
    sock = writer.transport.get_extra_info('socket')
    address = sock.getpeername()
    address = tuple(address[:2])
    itransport = TTransport(reader)
    iprotocol = protocol_cls(itransport)
    oprotocol = protocol_cls(writer)

    return ThriftConnection(service, iprot=iprotocol, oprot=oprotocol,
                            address=address, loop=loop, timeout=timeout)


class ThriftConnection:
    def __init__(self, service, *, iprot, oprot, address, loop=None, timeout=None):
        self.service = service
        self._reader = iprot.trans
        self._writer = oprot.trans
        self.loop = loop
        self.timeout = timeout
        self.address = address
        self.closed = False
        self._oprot = oprot
        self._iprot = iprot
        self._seqid = 0
        self._init_rpc_apis()

    def _init_rpc_apis(self):
        for api in self.service.thrift_services:
            setattr(self, api, functools.partial(self._send_call, api))

    @asyncio.coroutine
    def _send_call(self, api, *args, **kwargs):
        if self.closed:
            raise ConnectionClosedError('Connection closed')

        try:
            with async_timeout.timeout(self.timeout):
                _kw = args2kwargs(getattr(self.service, api + "_args").thrift_spec,
                                  *args)
                kwargs.update(_kw)
                result_cls = getattr(self.service, api + "_result")

                self._seqid += 1
                if self._seqid == sys.maxsize:
                    self._seqid = 0

                self._oprot.write_message_begin(api, TMessageType.CALL, self._seqid)
                args = getattr(self.service, api + '_args')()
                for k, v in kwargs.items():
                    setattr(args, k, v)
                args.write(self._oprot)
                self._oprot.write_message_end()
                try:
                    yield from self._oprot.trans.drain()
                except ConnectionError as e:
                    logger.debug('connection error {}'.format(str(e)))
                    raise ConnectionClosedError('the server has closed this connection')

                # writer.write
                # wait result only if non-oneway
                if not getattr(result_cls, "oneway"):
                    try:
                        result = yield from self._recv(api)
                    except TTransportException as e:
                        # handle EOF
                        if e.type == TTransportException.END_OF_FILE:
                            self.close()
                        raise
                    return result
        except asyncio.TimeoutError:
            self.close()
            raise

    @asyncio.coroutine
    def _recv(self, api):
        fname, mtype, rseqid = yield from self._iprot.read_message_begin()
        if rseqid != self._seqid:
            # transport should be closed if bad seq happened
            self.close()
            raise TApplicationException(TApplicationException.BAD_SEQUENCE_ID,
                                        fname + ' failed: out of sequence response')

        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            yield from self._iprot.read_struct(x)
            yield from self._iprot.read_message_end()
            raise x
        result = getattr(self.service, api + '_result')()
        yield from self._iprot.read_struct(result)
        yield from self._iprot.read_message_end()

        if hasattr(result, "success") and result.success is not None:
            return result.success

        # void api without throws
        if len(result.thrift_spec) == 0:
            return

        # check throws
        for k, v in result.__dict__.items():
            if k != "success" and v:
                raise v
        if hasattr(result, 'success'):
            raise TApplicationException(TApplicationException.MISSING_RESULT)

    def close(self):
        self._writer.close()
        self.closed = True



"""
The connection pool implementation is heavily borrowed from `aioredis`
"""
import asyncio
import collections
import sys

from .connection import create_connection
from .log import logger
from .util import async_task
from .errors import PoolClosedError

NotSet = object()

PY_35 = sys.version_info >= (3, 5)


@asyncio.coroutine
def create_pool(service, address, *, minsize=1, maxsize=10, loop=None, timeout=None):
    pool = ThriftPool(service, address, minsize=minsize,
                      maxsize=maxsize, loop=loop, timeout=timeout)
    try:
        yield from pool.fill_free(override_min=False)
    except Exception:
        pool.close()
        yield from pool.wait_closed()
        raise

    return pool


class ThriftPool:
    """Thrift connections pool.
    """

    def __init__(self, service, address,
                 *, minsize, maxsize, loop=None, timeout=None):
        assert isinstance(minsize, int) and minsize >= 0, (
            "minsize must be int >= 0", minsize, type(minsize))
        assert maxsize is not None, "Arbitrary pool size is disallowed."
        assert isinstance(maxsize, int) and maxsize > 0, (
            "maxsize must be int > 0", maxsize, type(maxsize))
        assert minsize <= maxsize, (
            "Invalid pool min/max sizes", minsize, maxsize)
        if loop is None:
            loop = asyncio.get_event_loop()
        self._address = address
        self._minsize = minsize
        self._loop = loop
        self._pool = collections.deque(maxlen=maxsize)
        self._used = set()
        self._acquiring = 0
        self._cond = asyncio.Condition(loop=loop)
        self._close_state = asyncio.Event(loop=loop)
        self._close_waiter = async_task(self._do_close(), loop=loop)
        self._service = service
        self._timeout = timeout

    @property
    def minsize(self):
        """Minimum pool size."""
        return self._minsize

    @property
    def maxsize(self):
        """Maximum pool size."""
        return self._pool.maxlen

    @property
    def size(self):
        """Current pool size."""
        return self.freesize + len(self._used) + self._acquiring

    @property
    def freesize(self):
        """Current number of free connections."""
        return len(self._pool)

    @asyncio.coroutine
    def clear(self):
        """Clear pool connections.

        Close and remove all free connections.
        this pattern is interesting
        """
        with (yield from self._cond):
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                waiters.append(conn.wait_closed())
            yield from asyncio.gather(*waiters, loop=self._loop)

    @asyncio.coroutine
    def _do_close(self):
        yield from self._close_state.wait()
        with (yield from self._cond):
            assert not self._acquiring, self._acquiring
            conn_num = 0
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                conn_num += 1
            for conn in self._used:
                conn.close()
                conn_num += 1
            logger.debug("Closed %d connections", conn_num)

    def close(self):
        """Close all free and in-progress connections and mark pool as closed.
        """
        if not self._close_state.is_set():
            self._close_state.set()

    @property
    def closed(self):
        """True if pool is closed."""
        return self._close_state.is_set()

    @asyncio.coroutine
    def wait_closed(self):
        """Wait until pool gets closed."""
        yield from asyncio.shield(self._close_waiter, loop=self._loop)

    @asyncio.coroutine
    def acquire(self):
        """Acquires a connection from free pool.

        Creates new connection if needed.
        """
        if self.closed:
            raise PoolClosedError("Pool is closed")
        with (yield from self._cond):
            if self.closed:
                raise PoolClosedError("Pool is closed")
            while True:
                yield from self.fill_free(override_min=True)
                if self.freesize:
                    conn = self._pool.popleft()
                    assert not conn.closed, conn
                    assert conn not in self._used, (conn, self._used)
                    self._used.add(conn)
                    return conn
                else:
                    # 没有可用的conn，等待release后 notify
                    yield from self._cond.wait()

    def release(self, conn):
        """Returns used connection back into pool.

        When returned connection has db index that differs from one in pool
        the connection will be closed and dropped.
        When queue of free connections is full the connection will be dropped.
        """
        assert conn in self._used, "Invalid connection, maybe from other pool"
        self._used.remove(conn)
        if not conn.closed:
            if self.maxsize and self.freesize < self.maxsize:
                self._pool.append(conn)
            else:
                # consider this connection as old and close it.
                conn.close()
        # FIXME: check event loop is not closed
        async_task(self._wakeup(), loop=self._loop)

    def _drop_closed(self):
        for i in range(self.freesize):
            conn = self._pool[0]
            if conn.closed:
                self._pool.popleft()
            else:
                self._pool.rotate(1)

    @asyncio.coroutine
    def fill_free(self, *, override_min):
        # drop closed connections first
        self._drop_closed()
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = yield from self._create_new_connection()
                self._pool.append(conn)
            finally:
                self._acquiring -= 1
                # connection may be closed at yield point
                self._drop_closed()
        # 还有可用的链接，返回
        if self.freesize:
            return
        if override_min:
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    conn = yield from self._create_new_connection()
                    self._pool.append(conn)
                finally:
                    self._acquiring -= 1
                    # connection may be closed at yield point
                    self._drop_closed()

    def _create_new_connection(self):
        return create_connection(self._service, self._address,
                                 loop=self._loop, timeout=self._timeout)

    @asyncio.coroutine
    def _wakeup(self, closing_conn=None):
        with (yield from self._cond):
            self._cond.notify()
        if closing_conn is not None:
            yield from closing_conn.wait_closed()

    def __enter__(self):
        raise RuntimeError(
            "'yield from' should be used as a context manager expression")

    def __exit__(self, *args):
        pass  # pragma: nocover

    def __iter__(self):
        # this method is needed to allow `yield`ing from pool
        conn = yield from self.acquire()
        return _ConnectionContextManager(self, conn)

    if PY_35:
        def __await__(self):
            # To make `with await pool` work
            conn = yield from self.acquire()
            return _ConnectionContextManager(self, conn)

        def get(self):
            '''Return async context manager for working with connection.

            async with pool.get() as conn:
                await conn.get(key)
            '''
            return _AsyncConnectionContextManager(self)


class _ConnectionContextManager:
    __slots__ = ('_pool', '_conn')

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        return self._conn

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None


if PY_35:
    class _AsyncConnectionContextManager:

        __slots__ = ('_pool', '_conn')

        def __init__(self, pool):
            self._pool = pool
            self._conn = None

        @asyncio.coroutine
        def __aenter__(self):
            self._conn = yield from self._pool.acquire()
            return self._conn

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_value, tb):
            try:
                self._pool.release(self._conn)
            finally:
                self._pool = None
                self._conn = None



import asyncio

from thriftpy.transport import TTransportException
import async_timeout

from .log import logger
from .processor import TProcessor
from .protocol import TBinaryProtocol
from .transport import TTransport


class Server:
    def __init__(self, processor, protocol_cls=TBinaryProtocol, timeout=None):
        self.processor = processor
        self.protocol_cls = protocol_cls
        self.timeout = timeout

    @asyncio.coroutine
    def __call__(self, reader, writer):
        itransport = TTransport(reader)
        iproto = self.protocol_cls(itransport)
        oproto = self.protocol_cls(writer)
        while not reader.at_eof():
            try:
                with async_timeout.timeout(self.timeout):
                    yield from self.processor.process(iproto, oproto)
            except TTransportException:
                logger.debug('transport exception')
                writer.close()
            except asyncio.TimeoutError:
                logger.debug('timeout when processing the client request')
                writer.close()
            except Exception:
                # app exception
                logger.exception('un handled app exception')
                writer.close()
        writer.close()


@asyncio.coroutine
def make_server(service, handler,
                host="localhost", port=9090,
                loop=None,
                protocol_cls=TBinaryProtocol,
                timeout=None
                ):
    processor = TProcessor(service, handler)
    if loop is None:
        loop = asyncio.get_event_loop()
    server = yield from asyncio.start_server(
        Server(processor, protocol_cls, timeout=timeout), host, port, loop=loop)
    return server
